@@ -1,7 +1,6 @@
 /* -*- mode: c; c-basic-offset: 4; indent-tabs-mode: nil -*- */
+/* plugins/kdb/ldap/libkdb_ldap/kdb_ldap.c */
 /*
- * lib/kdb/kdb_ldap/kdb_ldap.c
- *
  * Copyright (c) 2004-2005, Novell, Inc.
  * All rights reserved.
  *
@@ -115,12 +114,13 @@ krb5_ldap_read_startup_information(krb5_context context)
 
     SETUP_CONTEXT();
     if ((retval=krb5_ldap_read_krbcontainer_params(context, &(ldap_context->krbcontainer)))) {
-        prepend_err_str (context, "Unable to read Kerberos container", retval, retval);
+        prepend_err_str(context, _("Unable to read Kerberos container"),
+                        retval, retval);
         goto cleanup;
     }
 
     if ((retval=krb5_ldap_read_realm_params(context, context->default_realm, &(ldap_context->lrparams), &mask))) {
-        prepend_err_str (context, "Unable to read Realm", retval, retval);
+        prepend_err_str(context, _("Unable to read Realm"), retval, retval);
         goto cleanup;
     }
 
@@ -246,8 +246,8 @@ cleanup:
     return ret;
 }
 
-#define ERR_MSG1 "Unable to check if SASL EXTERNAL mechanism is supported by LDAP server. Proceeding anyway ..."
-#define ERR_MSG2 "SASL EXTERNAL mechanism not supported by LDAP server. Can't perform certificate-based bind."
+#define ERR_MSG1 _("Unable to check if SASL EXTERNAL mechanism is supported by LDAP server. Proceeding anyway ...")
+#define ERR_MSG2 _("SASL EXTERNAL mechanism not supported by LDAP server. Can't perform certificate-based bind.")
 
 /* Function to check if a LDAP server supports the SASL external mechanism
  *Return values:
@@ -329,12 +329,13 @@ krb5_ldap_open(krb5_context context, char *conf_section, char **db_args,
                 free (opt);
                 free (val);
                 status = EINVAL;
-                krb5_set_error_message (context, status, "'binddn' missing");
+                krb5_set_error_message(context, status, _("'binddn' missing"));
                 goto clean_n_exit;
             }
             if (val == NULL) {
                 status = EINVAL;
-                krb5_set_error_message (context, status, "'binddn' value missing");
+                krb5_set_error_message(context, status,
+                                       _("'binddn' value missing"));
                 free(opt);
                 goto clean_n_exit;
             }
@@ -350,12 +351,13 @@ krb5_ldap_open(krb5_context context, char *conf_section, char **db_args,
                 free (opt);
                 free (val);
                 status = EINVAL;
-                krb5_set_error_message (context, status, "'nconns' missing");
+                krb5_set_error_message(context, status, _("'nconns' missing"));
                 goto clean_n_exit;
             }
             if (val == NULL) {
                 status = EINVAL;
-                krb5_set_error_message (context, status, "'nconns' value missing");
+                krb5_set_error_message(context, status,
+                                       _("'nconns' value missing"));
                 free(opt);
                 goto clean_n_exit;
             }
@@ -365,12 +367,14 @@ krb5_ldap_open(krb5_context context, char *conf_section, char **db_args,
                 free (opt);
                 free (val);
                 status = EINVAL;
-                krb5_set_error_message (context, status, "'bindpwd' missing");
+                krb5_set_error_message(context, status,
+                                       _("'bindpwd' missing"));
                 goto clean_n_exit;
             }
             if (val == NULL) {
                 status = EINVAL;
-                krb5_set_error_message (context, status, "'bindpwd' value missing");
+                krb5_set_error_message(context, status,
+                                       _("'bindpwd' value missing"));
                 free(opt);
                 goto clean_n_exit;
             }
@@ -384,7 +388,8 @@ krb5_ldap_open(krb5_context context, char *conf_section, char **db_args,
         } else if (opt && !strcmp(opt, "host")) {
             if (val == NULL) {
                 status = EINVAL;
-                krb5_set_error_message (context, status, "'host' value missing");
+                krb5_set_error_message(context, status,
+                                       _("'host' value missing"));
                 free(opt);
                 goto clean_n_exit;
             }
@@ -417,37 +422,6 @@ krb5_ldap_open(krb5_context context, char *conf_section, char **db_args,
             }
 
             srv_cnt++;
-#ifdef HAVE_EDIRECTORY
-        } else if (opt && !strcmp(opt, "cert")) {
-            if (val == NULL) {
-                status = EINVAL;
-                krb5_set_error_message (context, status, "'cert' value missing");
-                free(opt);
-                goto clean_n_exit;
-            }
-
-            if (ldap_context->root_certificate_file == NULL) {
-                ldap_context->root_certificate_file = strdup(val);
-                if (ldap_context->root_certificate_file == NULL) {
-                    free (opt);
-                    free (val);
-                    status = ENOMEM;
-                    goto clean_n_exit;
-                }
-            } else {
-                char *newstr;
-
-                if (asprintf(&newstr, "%s %s",
-                             ldap_context->root_certificate_file, val) < 0) {
-                    free (opt);
-                    free (val);
-                    status = ENOMEM;
-                    goto clean_n_exit;
-                }
-                free(ldap_context->root_certificate_file);
-                ldap_context->root_certificate_file = newstr;
-            }
-#endif
         } else {
             /* ignore hash argument. Might have been passed from create */
             status = EINVAL;
@@ -456,10 +430,12 @@ krb5_ldap_open(krb5_context context, char *conf_section, char **db_args,
                  * temporary is passed in when kdb5_util load without -update is done.
                  * This is unsupported by the LDAP plugin.
                  */
-                krb5_set_error_message (context, status,
-                                        "open of LDAP directory aborted, plugin requires -update argument");
+                krb5_set_error_message(context, status,
+                                       _("open of LDAP directory aborted, "
+                                         "plugin requires -update argument"));
             } else {
-                krb5_set_error_message (context, status, "unknown option \'%s\'",
+                krb5_set_error_message (context, status,
+                                        _("unknown option \'%s\'"),
                                         opt?opt:val);
             }
             free(opt);
@@ -480,7 +456,8 @@ krb5_ldap_open(krb5_context context, char *conf_section, char **db_args,
             krb5_ldap_free_ldap_context(ldap_context);
         ldap_context = NULL;
         dal_handle->db_context = NULL;
-        prepend_err_str (context, "Error reading LDAP server params: ", status, status);
+        prepend_err_str(context, _("Error reading LDAP server params: "),
+                        status, status);
         goto clean_n_exit;
     }
     if ((status=krb5_ldap_db_init(context, ldap_context)) != 0) {
@@ -538,7 +515,7 @@ krb5_error_code
 krb5_ldap_check_policy_as(krb5_context kcontext, krb5_kdc_req *request,
                           krb5_db_entry *client, krb5_db_entry *server,
                           krb5_timestamp kdc_time, const char **status,
-                          krb5_data *e_data)
+                          krb5_pa_data ***e_data)
 {
     krb5_error_code retval;
 
